/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "GPUProcess.h"

#include "ArgumentCoders.h"
#include "Attachment.h"
#include "Decoder.h"
#include "GPUProcessConnectionInitializationParameters.h"
#include "GPUProcessConnectionParameters.h"
#include "GPUProcessCreationParameters.h"
#include "GPUProcessMessages.h"
#include "GPUProcessSessionParameters.h"
#include "HandleMessage.h"
#if ENABLE(MEDIA_STREAM)
#include "SandboxExtension.h"
#endif
#include "ShareableBitmap.h"
#include "WebCoreArgumentCoders.h"
#if HAVE(SC_CONTENT_SHARING_SESSION)
#include <WebCore/CaptureDevice.h>
#endif
#include <WebCore/MediaPlayerIdentifier.h>
#if ENABLE(MEDIA_STREAM)
#include <WebCore/MockMediaDevice.h>
#endif
#include <WebCore/ProcessIdentifier.h>
#if PLATFORM(MAC)
#include <WebCore/ScreenProperties.h>
#endif
#if ENABLE(MEDIA_STREAM)
#include <WebCore/SecurityOriginData.h>
#endif
#include <optional>
#include <pal/SessionID.h>
#if ENABLE(MEDIA_STREAM)
#include <wtf/Vector.h>
#endif
#if ENABLE(CFPREFS_DIRECT_MODE) || ENABLE(MEDIA_STREAM)
#include <wtf/text/WTFString.h>
#endif

namespace Messages {

namespace GPUProcess {

void CreateGPUConnectionToWebProcess::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<IPC::Attachment>&&, WebKit::GPUProcessConnectionInitializationParameters&&)>&& completionHandler)
{
    std::optional<std::optional<IPC::Attachment>> connectionIdentifier;
    decoder >> connectionIdentifier;
    if (!connectionIdentifier) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    std::optional<WebKit::GPUProcessConnectionInitializationParameters> parameters;
    decoder >> parameters;
    if (!parameters) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*connectionIdentifier), WTFMove(*parameters));
}

void CreateGPUConnectionToWebProcess::cancelReply(CompletionHandler<void(std::optional<IPC::Attachment>&&, WebKit::GPUProcessConnectionInitializationParameters&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<IPC::Attachment>>::create(), IPC::AsyncReplyError<WebKit::GPUProcessConnectionInitializationParameters>::create());
}

void CreateGPUConnectionToWebProcess::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<IPC::Attachment>& connectionIdentifier, const WebKit::GPUProcessConnectionInitializationParameters& parameters)
{
    encoder.get() << connectionIdentifier;
    encoder.get() << parameters;
    connection.sendSyncReply(WTFMove(encoder));
}

void PrepareToSuspend::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void PrepareToSuspend::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#if ENABLE(MEDIA_STREAM)

void UpdateCaptureAccess::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateCaptureAccess::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void UpdateCaptureAccess::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if HAVE(SC_CONTENT_SHARING_SESSION)

void ShowWindowPicker::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<WebCore::CaptureDevice>&&)>&& completionHandler)
{
    std::optional<std::optional<WebCore::CaptureDevice>> device;
    decoder >> device;
    if (!device) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*device));
}

void ShowWindowPicker::cancelReply(CompletionHandler<void(std::optional<WebCore::CaptureDevice>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<WebCore::CaptureDevice>>::create());
}

void ShowWindowPicker::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<WebCore::CaptureDevice>& device)
{
    encoder.get() << device;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

#if HAVE(SC_CONTENT_SHARING_SESSION)

void ShowScreenPicker::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<WebCore::CaptureDevice>&&)>&& completionHandler)
{
    std::optional<std::optional<WebCore::CaptureDevice>> device;
    decoder >> device;
    if (!device) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*device));
}

void ShowScreenPicker::cancelReply(CompletionHandler<void(std::optional<WebCore::CaptureDevice>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<WebCore::CaptureDevice>>::create());
}

void ShowScreenPicker::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<WebCore::CaptureDevice>& device)
{
    encoder.get() << device;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

void RequestBitmapImageForCurrentTime::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&& completionHandler)
{
    std::optional<WebKit::ShareableBitmap::Handle> handle;
    decoder >> handle;
    if (!handle) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*handle));
}

void RequestBitmapImageForCurrentTime::cancelReply(CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebKit::ShareableBitmap::Handle>::create());
}

void RequestBitmapImageForCurrentTime::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::ShareableBitmap::Handle& handle)
{
    encoder.get() << handle;
    connection.sendSyncReply(WTFMove(encoder));
}

void WebProcessConnectionCountForTesting::callReply(IPC::Decoder& decoder, CompletionHandler<void(uint64_t&&)>&& completionHandler)
{
    std::optional<uint64_t> count;
    decoder >> count;
    if (!count) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*count));
}

void WebProcessConnectionCountForTesting::cancelReply(CompletionHandler<void(uint64_t&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<uint64_t>::create());
}

void WebProcessConnectionCountForTesting::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t count)
{
    encoder.get() << count;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace GPUProcess

} // namespace Messages

namespace WebKit {

void GPUProcess::didReceiveGPUProcessMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::GPUProcess::InitializeGPUProcess::name())
        return IPC::handleMessage<Messages::GPUProcess::InitializeGPUProcess>(connection, decoder, this, &GPUProcess::initializeGPUProcess);
    if (decoder.messageName() == Messages::GPUProcess::CreateGPUConnectionToWebProcess::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::CreateGPUConnectionToWebProcess>(connection, decoder, this, &GPUProcess::createGPUConnectionToWebProcess);
    if (decoder.messageName() == Messages::GPUProcess::PrepareToSuspend::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::PrepareToSuspend>(connection, decoder, this, &GPUProcess::prepareToSuspend);
    if (decoder.messageName() == Messages::GPUProcess::ProcessDidResume::name())
        return IPC::handleMessage<Messages::GPUProcess::ProcessDidResume>(connection, decoder, this, &GPUProcess::processDidResume);
    if (decoder.messageName() == Messages::GPUProcess::AddSession::name())
        return IPC::handleMessage<Messages::GPUProcess::AddSession>(connection, decoder, this, &GPUProcess::addSession);
    if (decoder.messageName() == Messages::GPUProcess::RemoveSession::name())
        return IPC::handleMessage<Messages::GPUProcess::RemoveSession>(connection, decoder, this, &GPUProcess::removeSession);
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::SetMockCaptureDevicesEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetMockCaptureDevicesEnabled>(connection, decoder, this, &GPUProcess::setMockCaptureDevicesEnabled);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::SetOrientationForMediaCapture::name())
        return IPC::handleMessage<Messages::GPUProcess::SetOrientationForMediaCapture>(connection, decoder, this, &GPUProcess::setOrientationForMediaCapture);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::UpdateSandboxAccess::name())
        return IPC::handleMessage<Messages::GPUProcess::UpdateSandboxAccess>(connection, decoder, this, &GPUProcess::updateSandboxAccess);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::UpdateCaptureAccess::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::UpdateCaptureAccess>(connection, decoder, this, &GPUProcess::updateCaptureAccess);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::UpdateCaptureOrigin::name())
        return IPC::handleMessage<Messages::GPUProcess::UpdateCaptureOrigin>(connection, decoder, this, &GPUProcess::updateCaptureOrigin);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::AddMockMediaDevice::name())
        return IPC::handleMessage<Messages::GPUProcess::AddMockMediaDevice>(connection, decoder, this, &GPUProcess::addMockMediaDevice);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::ClearMockMediaDevices::name())
        return IPC::handleMessage<Messages::GPUProcess::ClearMockMediaDevices>(connection, decoder, this, &GPUProcess::clearMockMediaDevices);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::RemoveMockMediaDevice::name())
        return IPC::handleMessage<Messages::GPUProcess::RemoveMockMediaDevice>(connection, decoder, this, &GPUProcess::removeMockMediaDevice);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::ResetMockMediaDevices::name())
        return IPC::handleMessage<Messages::GPUProcess::ResetMockMediaDevices>(connection, decoder, this, &GPUProcess::resetMockMediaDevices);
#endif
#if ENABLE(MEDIA_STREAM)
    if (decoder.messageName() == Messages::GPUProcess::SetMockCaptureDevicesInterrupted::name())
        return IPC::handleMessage<Messages::GPUProcess::SetMockCaptureDevicesInterrupted>(connection, decoder, this, &GPUProcess::setMockCaptureDevicesInterrupted);
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::GPUProcess::DisplayConfigurationChanged::name())
        return IPC::handleMessage<Messages::GPUProcess::DisplayConfigurationChanged>(connection, decoder, this, &GPUProcess::displayConfigurationChanged);
#endif
#if PLATFORM(MAC)
    if (decoder.messageName() == Messages::GPUProcess::SetScreenProperties::name())
        return IPC::handleMessage<Messages::GPUProcess::SetScreenProperties>(connection, decoder, this, &GPUProcess::setScreenProperties);
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    if (decoder.messageName() == Messages::GPUProcess::ShowWindowPicker::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::ShowWindowPicker>(connection, decoder, this, &GPUProcess::showWindowPicker);
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    if (decoder.messageName() == Messages::GPUProcess::ShowScreenPicker::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::ShowScreenPicker>(connection, decoder, this, &GPUProcess::showScreenPicker);
#endif
#if ENABLE(MEDIA_SOURCE)
    if (decoder.messageName() == Messages::GPUProcess::SetWebMParserEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetWebMParserEnabled>(connection, decoder, this, &GPUProcess::setWebMParserEnabled);
#endif
#if ENABLE(WEBM_FORMAT_READER)
    if (decoder.messageName() == Messages::GPUProcess::SetWebMFormatReaderEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetWebMFormatReaderEnabled>(connection, decoder, this, &GPUProcess::setWebMFormatReaderEnabled);
#endif
#if ENABLE(OPUS)
    if (decoder.messageName() == Messages::GPUProcess::SetOpusDecoderEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetOpusDecoderEnabled>(connection, decoder, this, &GPUProcess::setOpusDecoderEnabled);
#endif
#if ENABLE(VORBIS)
    if (decoder.messageName() == Messages::GPUProcess::SetVorbisDecoderEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetVorbisDecoderEnabled>(connection, decoder, this, &GPUProcess::setVorbisDecoderEnabled);
#endif
#if ENABLE(MEDIA_SOURCE) && HAVE(AVSAMPLEBUFFERVIDEOOUTPUT)
    if (decoder.messageName() == Messages::GPUProcess::SetMediaSourceInlinePaintingEnabled::name())
        return IPC::handleMessage<Messages::GPUProcess::SetMediaSourceInlinePaintingEnabled>(connection, decoder, this, &GPUProcess::setMediaSourceInlinePaintingEnabled);
#endif
#if HAVE(SCREEN_CAPTURE_KIT)
    if (decoder.messageName() == Messages::GPUProcess::SetUseScreenCaptureKit::name())
        return IPC::handleMessage<Messages::GPUProcess::SetUseScreenCaptureKit>(connection, decoder, this, &GPUProcess::setUseScreenCaptureKit);
#endif
    if (decoder.messageName() == Messages::GPUProcess::RequestBitmapImageForCurrentTime::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::RequestBitmapImageForCurrentTime>(connection, decoder, this, &GPUProcess::requestBitmapImageForCurrentTime);
#if ENABLE(CFPREFS_DIRECT_MODE)
    if (decoder.messageName() == Messages::GPUProcess::NotifyPreferencesChanged::name())
        return IPC::handleMessage<Messages::GPUProcess::NotifyPreferencesChanged>(connection, decoder, this, &GPUProcess::notifyPreferencesChanged);
#endif
    if (decoder.messageName() == Messages::GPUProcess::WebProcessConnectionCountForTesting::name())
        return IPC::handleMessageAsync<Messages::GPUProcess::WebProcessConnectionCountForTesting>(connection, decoder, this, &GPUProcess::webProcessConnectionCountForTesting);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
